@@ -1,6 +1,6 @@
 """Basic Backend Interface."""
 import logging
-from typing import Any, Union, List, Optional, Callable
+from typing import Any, Union, List, Optional, Callable, Set
 
 import numpy as np
 
@@ -11,6 +11,9 @@ class Backend:
     """Basic Interface for a Moving Targets Master Backend."""
 
     _LOGGER: logging.Logger = logging.getLogger('Backend')
+    """The inner backend logger instance."""
+
+    _INDICATORS: Set[str] = {'>', '>=', '<', '<=', '==', '!='}
 
     @staticmethod
     def _transposed_axes(dim: int, index: int, place: int) -> List[int]:
@@ -66,13 +69,13 @@ class Backend:
         :param msg:
             The warning explanation.
         """
-        if not (exp is None and aux is None) and exp != aux:
+        if aux != 'auto' and exp != aux:
             cls._LOGGER.warning(f"'aux={aux}' has no effect since the solver needs {exp} auxiliary variables to {msg}.")
 
     def __init__(self, sum_fn: Callable = lambda v: np.sum(v)):
         """
         :param sum_fn:
-            The backend function to perform a sum over a one-dimensional vector of model variables.
+            The backend function to perform a sum over a one-dimensional array of model variables.
         """
         super(Backend, self).__init__()
 
@@ -155,33 +158,35 @@ class Backend:
         """
         raise NotImplementedError(not_implemented_message(name='maximize'))
 
-    def add_constraints(self, constraints: Union[List, np.ndarray], name: Optional[str] = None) -> Any:
-        """Adds a list of constraints to the model.
-
-        :param constraints:
-            The list of constraints.
-
-        :param name:
-            The constraints name.
+    def get_objective(self) -> float:
+        """Gets the objective value of the solved model.
 
         :return:
-            The backend itself.
+            The objective value.
         """
-        raise NotImplementedError(not_implemented_message(name='add_constraints'))
+        raise NotImplementedError(not_implemented_message(name='get_objective'))
 
-    def add_constraint(self, constraint, name: Optional[str] = None) -> Any:
-        """Adds a constraint to the model.
+    def get_values(self, expressions: np.ndarray) -> np.ndarray:
+        """Gets the values of an array of expressions as found in the model solution.
 
-        :param constraint:
-            The constraint.
-
-        :param name:
-            The constraint name.
+        :param expressions:
+            The array of expressions.
 
         :return:
-            The backend itself.
+            The array of solution values.
         """
-        return self.add_constraints(constraints=[constraint], name=name)
+        raise NotImplementedError(not_implemented_message(name='get_values'))
+
+    def get_value(self, expression: Any) -> float:
+        """Gets the value of an expressions as found in the model solution.
+
+        :param expression:
+            The expressions.
+
+        :return:
+            The solution value.
+        """
+        return self.get_values(np.array([expression]))[0]
 
     def add_binary_variables(self, *keys: int, name: Optional[str] = None) -> np.ndarray:
         """Creates an array of binary model variables.
@@ -381,36 +386,6 @@ class Backend:
         names = np.array(self._nested_names(*keys, name=name)).flatten()
         return np.reshape([self.add_variable(vtype=vtype, lb=lb, ub=ub, name=n) for n in names], keys)
 
-    def get_objective(self) -> float:
-        """Gets the objective value of the solved model.
-
-        :return:
-            The objective value.
-        """
-        raise NotImplementedError(not_implemented_message(name='get_objective'))
-
-    def get_values(self, expressions: np.ndarray) -> np.ndarray:
-        """Gets the values of a vector of expressions as found in the model solution.
-
-        :param expressions:
-            The vector of expressions.
-
-        :return:
-            The vector of solution values.
-        """
-        raise NotImplementedError(not_implemented_message(name='get_values'))
-
-    def get_value(self, expression: Any) -> float:
-        """Gets the value of an expressions as found in the model solution.
-
-        :param expression:
-            The expressions.
-
-        :return:
-            The solution value.
-        """
-        return self.get_values(np.array([expression]))[0]
-
     def aux(self,
             expressions: Any,
             aux_vtype: Optional[str] = 'continuous',
@@ -418,13 +393,13 @@ class Backend:
             aux_ub: float = float('inf'),
             aux_name: Optional[str] = None) -> Any:
         """If the 'aux_vtype' parameter is None, it simply return the input expressions, otherwise it builds and return
-        an auxiliary variable (or a vector of variables) having the given vtype and the respective other properties,
-        which is equal to the given expression (or vector of expressions). Using auxiliary variables may come in handy
-        when dealing with huge datasets since they can considerably speedup the model formulation; still, imposing
-        equality constraints on certain expressions may lead to solving errors due to broken model assumptions.
+        an auxiliary variable (or an array of variables) having the given vtype and the respective other properties,
+        which is equal to the given expression (or an array of expressions). Using auxiliary variables may come in
+        handy when dealing with huge datasets since they can considerably speedup the model formulation; still,
+        imposing equality constraints on certain expressions may lead to solving errors due to broken model assumptions.
 
         :param expressions:
-            Either a single expression or a vector of expressions.
+            Either a single expression or a array of expressions.
 
         :param aux_vtype:
             Either None (or 'auto') or the auxiliary variables type, usually 'binary', 'integer', or 'continuous'.
@@ -452,7 +427,159 @@ class Backend:
             self.add_constraints([v == e for v, e in zip(variables.flatten(), expressions.flatten())])
             return variables if is_numpy else variables[0]
 
-    def sum(self, a: np.ndarray, axis: Optional[int] = None, asarray: bool = False, aux: Optional[str] = None) -> Any:
+    def add_constraints(self, constraints: Union[List, np.ndarray], name: Optional[str] = None) -> Any:
+        """Adds a list of constraints to the model.
+
+        :param constraints:
+            The list of constraints.
+
+        :param name:
+            The constraints name.
+
+        :return:
+            The backend itself.
+        """
+        constraints = np.array(constraints)
+        names = np.array(self._nested_names(*constraints.shape, name=name))
+        for c, n in zip(constraints.flatten(), names.flatten()):
+            self.add_constraint(constraint=c, name=n)
+        return self
+
+    def add_constraint(self, constraint, name: Optional[str] = None) -> Any:
+        """Adds a constraint to the model.
+
+        :param constraint:
+            The constraint.
+
+        :param name:
+            The constraint name.
+
+        :return:
+            The backend itself.
+        """
+        raise NotImplementedError(not_implemented_message(name='add_constraint'))
+
+    def add_indicator_constraints(self,
+                                  indicators: np.ndarray,
+                                  expressions: Union[List, np.ndarray],
+                                  value: int = 1,
+                                  name: Optional[str] = None) -> Any:
+        """Impose a set of indicator constraints over the given expressions using the given binary indicators. The
+        indicators indicators[i] are such that if indicators[i] == value (with value either 0 or 1), then the i-th
+        expression holds, but in case indicators[i] == not value, then there is no information about that expression.
+
+        :param indicators:
+            An array of binary variables on which to impose the indicator constraints.
+
+        :param expressions:
+            An array or list of expressions for which it is necessary to check whether they hold or not.
+
+        :param value:
+            The value assumed by a binary variable when the respective expression holds, either 0 or 1.
+
+        :param name:
+            The constraints name.
+
+        :return:
+            The backend itself.
+
+        :raise `BackendError`:
+            If the backend cannot handle indicator variables.
+        """
+        expressions = np.array(expressions)
+        names = np.array(self._nested_names(*expressions.shape, name=name))
+        for i, e, n in zip(indicators.flatten(), expressions.flatten(), names.flatten()):
+            self.add_indicator_constraint(indicator=i, expression=e, value=value, name=n)
+        return self
+
+    def add_indicator_constraint(self,
+                                 indicator: Any,
+                                 expression: Any,
+                                 value: int = 1,
+                                 name: Optional[str] = None) -> Any:
+        """Impose an indicator constraint over the given expression using the given binary indicator. The indicator
+        is such that if indicator == value (with value either 0 or 1), then the expression holds, but in case
+        indicator == not value, then there is no information about the expression.
+
+        :param indicator:
+            The binary variable on which to impose the indicator constraints.
+
+        :param expression:
+            The expression for which it is necessary to check whether they hold or not.
+
+        :param value:
+            The value assumed by a binary variable when the respective expression holds, either 0 or 1.
+
+        :param name:
+            The constraint name.
+
+        :return:
+            The backend itself.
+
+        :raise `BackendError`:
+            If the backend cannot handle indicator variables.
+        """
+        raise BackendError(unsupported='indicator constraints')
+
+    def is_greater(self, a: np.ndarray, b: Union[float, np.ndarray]) -> np.ndarray:
+        """Builds auxiliary binary indicator variables which take value one if the expressions in the first array are
+         greater than the expressions in the second array. Please note that this is enforced via indicator constraints
+         so that if z[i] == 1 -> a[i] >= b[i] and if z[i] == 0 -> a[i] <= b[i], thus in case a[i] is strictly equal to
+         b[i] the indicator variable z[i] can assume both values.
+
+        :param a:
+            The first array representing the left-hand sides.
+
+        :param b:
+            Either a single reference value or a second array of expressions representing the right-hand sides.
+
+        :return:
+            The array of binary indicator variables.
+
+        :raise `BackendError`:
+            If the backend cannot handle indicator variables.
+        """
+        # if b is an array of expressions, the i-th right hand side will be b[i] (with b flattened), otherwise it is b
+        if isinstance(b, np.ndarray):
+            b = b.flatten()
+            rhs = lambda i: b[i]
+        else:
+            rhs = lambda i: b
+        z = self.add_binary_variables(a.size)
+        self.add_indicator_constraints(z, expressions=[lhs >= rhs(i) for i, lhs in enumerate(a.flatten())], value=1)
+        self.add_indicator_constraints(z, expressions=[lhs <= rhs(i) for i, lhs in enumerate(a.flatten())], value=0)
+        return z.reshape(a.shape)
+
+    def is_less(self, a: np.ndarray, b: np.ndarray) -> np.ndarray:
+        """Builds auxiliary binary indicator variables which take value one if the expressions in the first array are
+         lower than the expressions in the second array. Please note that this is enforced via indicator constraints
+         so that if z[i] == 1 -> a[i] <= b[i] and if z[i] == 0 -> a[i] >= b[i], thus in case a[i] is strictly equal to
+         b[i] the indicator variable z[i] can assume both values.
+
+        :param a:
+            The first array.
+
+        :param b:
+            The second array.
+
+        :return:
+            The array of binary indicator variables.
+
+        :raise `BackendError`:
+            If the backend cannot handle indicator variables.
+        """
+        # if b is an array of expressions, the i-th right hand side will be b[i] (with b flattened), otherwise it is b
+        if isinstance(b, np.ndarray):
+            b = b.flatten()
+            rhs = lambda i: b[i]
+        else:
+            rhs = lambda i: b
+        z = self.add_binary_variables(a.size)
+        self.add_indicator_constraints(z, expressions=[lhs <= rhs(i) for i, lhs in enumerate(a.flatten())], value=1)
+        self.add_indicator_constraints(z, expressions=[lhs >= rhs(i) for i, lhs in enumerate(a.flatten())], value=0)
+        return z.reshape(a.shape)
+
+    def sum(self, a: np.ndarray, axis: Optional[int] = None, asarray: bool = False, aux: Optional[str] = 'auto') -> Any:
         """Computes the sum of an array of variables.
 
         :param a:
@@ -462,8 +589,8 @@ class Backend:
             The dimension on which to aggregate or None to aggregate the whole data.
 
         :param asarray:
-            In case the aggregation should return a single expression, whether to return it as a numpy zero-dimensional
-            array or as the expression itself.
+            In case the aggregation returns a single expression, whether to return it as a zero-dimensional numpy array
+            or as the expression itself.
 
         :param aux:
             The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
@@ -499,7 +626,7 @@ class Backend:
             expressions = np.reshape(expressions, new_shape) if len(new_shape) > 0 or asarray else expressions[0]
         return self.aux(expressions, aux_vtype=aux)
 
-    def square(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
+    def square(self, a: np.ndarray, aux: Optional[str] = 'auto') -> np.ndarray:
         """Computes the squared values over an array of variables.
 
         :param a:
@@ -515,14 +642,14 @@ class Backend:
             constraints on certain expressions may lead to solving errors due to broken model assumptions.
 
         :return:
-            The vector of squared values.
+            The array of squared values.
 
         :raise `BackendError`:
             If the backend cannot handle squared values.
         """
         raise BackendError(unsupported='squared values')
 
-    def sqrt(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
+    def sqrt(self, a: np.ndarray, aux: Optional[str] = 'auto') -> np.ndarray:
         """Computes the squared roots over an array of variables.
 
         :param a:
@@ -538,12 +665,12 @@ class Backend:
             constraints on certain expressions may lead to solving errors due to broken model assumptions.
 
         :return:
-            The vector of squared roots.
+            The array of squared roots.
         """
         raise BackendError(unsupported='squared roots')
 
-    def abs(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
-        """Computes the absolute values over a vector of variables.
+    def abs(self, a: np.ndarray, aux: Optional[str] = 'auto') -> np.ndarray:
+        """Computes the absolute values over an array of variables.
 
         :param a:
             An array of model variables.
@@ -558,14 +685,14 @@ class Backend:
             constraints on certain expressions may lead to solving errors due to broken model assumptions.
 
         :return:
-            The vector of absolute values.
+            The array of absolute values.
 
         :raise `BackendError`:
             If the backend cannot handle absolute values.
         """
         raise BackendError(unsupported='absolute values')
 
-    def log(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
+    def log(self, a: np.ndarray, aux: Optional[str] = 'auto') -> np.ndarray:
         """Computes the logarithms over an array of variables.
 
         :param a:
@@ -581,14 +708,18 @@ class Backend:
             constraints on certain expressions may lead to solving errors due to broken model assumptions.
 
         :return:
-            The vector of logarithms.
+            The array of logarithms.
 
         :raise `BackendError`:
             If the backend cannot handle logarithms.
         """
         raise BackendError(unsupported='logarithms')
 
-    def mean(self, a: np.ndarray, axis: Optional[int] = None, asarray: bool = False, aux: Optional[str] = None) -> Any:
+    def mean(self,
+             a: np.ndarray,
+             axis: Optional[int] = None,
+             asarray: bool = False,
+             aux: Optional[str] = 'auto') -> Any:
         """Computes the mean of an array of variables.
 
         :param a:
@@ -598,8 +729,8 @@ class Backend:
             The dimension on which to aggregate or None to aggregate the whole data.
 
         :param asarray:
-            In case the aggregation should return a single expression, whether to return it as a numpy zero-dimensional
-            array or as the expression itself.
+            In case the aggregation returns a single expression, whether to return it as a zero-dimensional numpy array
+            or as the expression itself.
 
         :param aux:
             The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
@@ -627,8 +758,8 @@ class Backend:
             The dimension on which to aggregate or None to aggregate the whole data.
 
         :param asarray:
-            In case the aggregation should return a single expression, whether to return it as a numpy zero-dimensional
-            array or as the expression itself.
+            In case the aggregation returns a single expression, whether to return it as a zero-dimensional numpy array
+            or as the expression itself.
 
         :param aux:
             The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
@@ -667,7 +798,7 @@ class Backend:
         squared_expressions = self.square(diff_expressions, aux=aux_squared)
         return self.mean(squared_expressions, axis=axis, asarray=asarray, aux=aux)
 
-    def add(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = None):
+    def add(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = 'auto'):
         """Performs the pairwise sum between two arrays.
 
         :param a:
@@ -690,7 +821,7 @@ class Backend:
         """
         return self.aux(expressions=a + b, aux_vtype=aux)
 
-    def subtract(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = None):
+    def subtract(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = 'auto'):
         """Performs the pairwise subtraction between two arrays.
 
         :param a:
@@ -713,7 +844,7 @@ class Backend:
         """
         return self.aux(expressions=a - b, aux_vtype=aux)
 
-    def multiply(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = None):
+    def multiply(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = 'auto'):
         """Performs the pairwise product between two arrays.
 
         :param a:
@@ -736,7 +867,7 @@ class Backend:
         """
         return self.aux(expressions=a * b, aux_vtype=aux)
 
-    def divide(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = None):
+    def divide(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = 'auto'):
         """Performs the pairwise division between two arrays.
 
         :param a:
@@ -759,8 +890,12 @@ class Backend:
         """
         return self.aux(expressions=a / b, aux_vtype=aux)
 
-    def dot(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = None) -> Union[Any, np.ndarray]:
-        """Performs the vector product between two arrays.
+    def dot(self,
+            a: np.ndarray,
+            b: np.ndarray,
+            asarray: bool = False,
+            aux: Optional[str] = 'auto') -> Union[Any, np.ndarray]:
+        """Performs the dot product between two arrays.
 
         :param a:
             The first array (at most 2d).
@@ -768,6 +903,10 @@ class Backend:
         :param b:
             The second array (at most 2d).
 
+        :param asarray:
+            In case the aggregation returns a single expression, whether to return it as a zero-dimensional numpy array
+            or as the expression itself.
+
         :param aux:
             The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
             the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
@@ -780,14 +919,5 @@ class Backend:
         :return:
             The dot product between a and b.
         """
-        a = a if a.ndim == 2 else a.reshape((1, len(a)))
-        b = b if b.ndim == 2 else b.reshape((len(b), 1))
-        # we use a list of lists instead of an array to let numpy handle the data type on its own
-        r = []
-        for i, row in enumerate(a):
-            r.append([])
-            for j, column in enumerate(b.T):
-                pairwise_products = self.multiply(a=row, b=column, aux=aux)
-                r[i].append(self.sum(pairwise_products, aux=aux))
-        r = np.array(r)
-        return r[0, 0] if r.size == 1 else r.squeeze()
+        dot_expression = self.aux(expressions=a @ b, aux_vtype=aux)
+        return np.array(dot_expression) if asarray else dot_expression
